/*
This file is part of Youga (Youpi Gamepad Front-end)
Copyright (C) 2017 AlphaK - http://www.alphak.net/

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

var version = '1.0';
var baseURL = 'ws://localhost:18080/';
var isRotation = false;
var isConnected = false;
var STATUS_DISCONNECTED = 'Status: not connected'
var STATUS_CONNECTED = 'Status: connected'
var websocket = null;

var defaultGamepadId = 0;
var A = 'A';
var B = 'B';
var gamepadMap = [];
var oldInstructionSet = '';
var configOK = false;
var emptyInstructionSet = 'C!W!A,N,U,0,1';
var backToInitialInstructionSet = 'C!W!A,I,H,1,1';
var appendPositionInstructionSet = 'I!!';
var replayInstructionSet = '*** SCRIPT REPLAY ***';

// Map of human readable settings to Youba instruction parts
var configMap = {
	'baseLeft': [0, 'L', null], 
	'baseRight': [0, 'R', null], 
	'shoulderDown': [1, 'L', null], 
	'shoulderUp': [1, 'R', null], 
	'elbowDown': [2, 'R', null], 
	'elbowUp': [2, 'L', null], 
	'wristDown': [3, 'L', null], 
	'wristUp': [3, 'R', null], 
	'handLeft': [4, 'L', null], 
	'handRight': [4, 'R', null], 
	'pliersTight': [5, 'L', null], 
	'pliersLoose': [5, 'R', null],
	'backToInitial': [null, null, 1],
	'appendPosition': [null, null, 2],
	'replayScript': [null, null, 3]
};

// Initialize a table for motor values
function newMotorTable(param) {
	return [param, param, param, param, param, param] ;
}

// Send a command to the back-end
function sendYoupi(data) {
	if(isConnected && configOK) {
		websocket.send(data);
	}
}

// Triggered when the websocket is connected
function wsOnOpen(evt) {
	updateStatusAndButtons(true);
}

// Triggered when the websocket is disconnected
function wsOnClose(evt) {
	updateStatusAndButtons(false);
}

// Triggered when the websocket has errors
function wsOnError(evt) {
	websocket.close();
	updateStatusAndButtons(false);
}

// Triggered when the websocket sends data
function wsOnMessage(evt) {
	if(evt.data != null) {
		if(evt.data != 'OK') {
			if (evt.data.length > 3 && evt.data.substring(0,3) == 'OK ') {
				appendPositionToRecap(evt.data.substring(3));
			}
		}
	}
}

// Polls the robot for current position and append it to the bottom script
function appendPositionToRecap(position) {
	var arrayPos = JSON.parse(position);
	var currentRecap = $('#recap').text();
	var appendText = '\nA!W!';
	for(var i=0 ; i<arrayPos.length ; i++) {
		if(i != 0) appendText += '|';
		appendText += (i+',A,H,'+arrayPos[i]+',1');
	}
	$('#recap').text(currentRecap + appendText);
}

// Updates the buttons and status according to the state of the websocket
function updateStatusAndButtons(connected) {
	isConnected = connected;
	$('#status').text(connected ? STATUS_CONNECTED : STATUS_DISCONNECTED);
	$('#url').prop('disabled', connected);
	$('#btnConnect').prop('disabled', connected);
	$('#btnDisconnect').prop('disabled', !connected);
}

// Update the memory object representation of the instructionSet according to gamepad current values
function updateInstructionSetObjectAndGui() {
	gid = (typeof gamepadId == 'undefined') ? defaultGamepadId : gamepadId;
	if (gamepadSupport.gamepads[gid]) {
	
		// We have a recognized gamepad
		var gamepad = gamepadSupport.gamepads[gid];
	
		// Start with empty objects		
		var instructionTable = newMotorTable('');
		var speedTable = newMotorTable('0%');
		var dirTable = newMotorTable('-');
		
		// Iterate into gamepadMap (object built at startup from config file)
		for(var key in gamepadMap) {
			var map = gamepadMap[key];
			var iType = map['type'];
			var iIndex = map['index'];
			var iSign = map['sign'];
			var gpValue = null;
			// Get axis/button value
			if(iType == A && typeof gamepad.axes != 'undefined' && iIndex < gamepad.axes.length) {
				gpValue = gamepad.axes[iIndex];
			} else if(iType == B && typeof gamepad.buttons != 'undefined' && iIndex < gamepad.buttons.length) {
				gpValue = gamepad.buttons[iIndex].value;
			}
			// Do the complex calculations only if gamepad axis/button value is known
			if(gpValue !== null) {
				map['value'] = adjustGamepadValue(gpValue); // Fine tune value (between -1.0 and 1.0 with one decimal)
				
				// Do the complex calculations only if gamepad adjusted axis/button value is not 0
				if(map['value'] != 0) {
					var sign = (iSign) ? 1 : -1;
					
					// Do the complex calculations only the sign of the axis/button value matches the sign in the config file
					if(Math.sign(sign) == Math.sign(map['value'])) {
					
						// In case it is a special command, handle it in a dedicated function and don't bother checking other axes/buttons
						if(map['youbaSpecial'] != null) {
							handleSpecialCommand(map);
							return;
						}
					
						// It is not a special command, handle it normally
						var direction = map['youbaDir']; // Get direction from the config file
						var period = Math.round(1 / Math.abs(map['value']) * 100) / 100; // Get period from gamepad axis/button value, standardized value between 1(fastest) and 10(slowest)
						instructionTable[map['youbaMotor']] = map['youbaMotor'] + ',' + direction + ',U,0,' + period ; // Build instruction from previously computed variables
						speedTable[map['youbaMotor']] = Math.round(1 / period * 100) + '%'; // Compute displayed speed from period, between 10%(slowest) and 100%(fastest)
						dirTable[map['youbaMotor']] = direction; // Get direction from previously computed variable
						// Now we have 3 different functional arrays whose index corresponding to the motor ID is up to date
					}
				}
			}
		}
		// *** Now we have 3 different functional arrays with all indexes up to date ***
		
		// Compute the instructionSet as a string and print it in the debug placeholder
		var curInstructionSet = buildInstructionSet(instructionTable);
		$('#debug').val(curInstructionSet);
		
		// If instructionSet changed from last iteration, update the graphical user interface and send the instructionSet to Youba
		if(curInstructionSet != oldInstructionSet) {
			updateFrontendDisplay(instructionTable, speedTable, dirTable);
			
			// Ensure that if no button is pressed, back to initial and replay script commands are still executed
			if(!(curInstructionSet == emptyInstructionSet && (oldInstructionSet == backToInitialInstructionSet || oldInstructionSet == replayInstructionSet))) {
				sendYoupi(curInstructionSet);
			}
			oldInstructionSet = curInstructionSet;
		}
	}
}

// Handle special commands (back to initial position, get current position, replay script)
function handleSpecialCommand(pMap) {
	var curInstructionSet = emptyInstructionSet;
	if(pMap['youbaSpecial'] == 1 || pMap['youbaSpecial'] == 2) {
		curInstructionSet = (pMap['youbaSpecial'] == 1) ? backToInitialInstructionSet : appendPositionInstructionSet;
		$('#debug').val(curInstructionSet);
		if(curInstructionSet != oldInstructionSet) {
			sendYoupi(curInstructionSet);
			oldInstructionSet = curInstructionSet;
		}
	} else if (pMap['youbaSpecial'] == 3) {
		curInstructionSet = replayInstructionSet;
		$('#debug').val(curInstructionSet);
		if(curInstructionSet != oldInstructionSet) {
			replayCurrentScript();
			oldInstructionSet = curInstructionSet;
		}
	}
}

// Get the replay script content and resend it to the backend
function replayCurrentScript() {
	var i;
	var currentRecap = $('#recap').text();
	var lines = currentRecap.split('\n');
	for(i=0; i<lines.length; i++) {
		sendYoupi(lines[i]);
	}
}

// Fine tune the gamepad axis/button value (stick to 0% or 100% if close enough)
function adjustGamepadValue (pValue) {
	if(pValue > 0.96) { pValue = 1; }
	if(pValue < -0.96) { pValue = -1; }
	if(pValue > 0 && pValue < 0.15) { pValue = 0; }
	if(pValue < 0 && pValue > -0.15) { pValue = 0; }
	// Return a standardized value between -1.0 to 1.0 with only one decimal
	return Math.sign(pValue) * Math.floor(Math.abs(pValue) * 10) / 10;
}

// Update the front-end display (graphical robot, arrows and speed percentage)
function updateFrontendDisplay (pInstructionTable, pSpeedTable, pDirTable) {
	// Hide all layers
	$('img.img-robot-layer').hide();
	$('img.img-arrow-layer').hide();
	$('tr td.tdSpeed').text('0%');
	// Show specific layer and value according to instructionChain
	for(var i = 0 ; i < pInstructionTable.length ; i++) {
		if(pInstructionTable[i] != '') {
			$('#img-m' + i).show();
			$('tr#m' + i + ' td.' + i + pDirTable[i] + ' img.img-arrow-layer').show();
			$('tr#m' + i + ' td.tdSpeed').text(pSpeedTable[i]);
		}
	}
}

// Build the instructionSet as a string that can be sent to Youba
function buildInstructionSet(pInstructionTable) {
	var tmpInstructionSet = '';
	for (var i=0 ; i < pInstructionTable.length ; i++) {
		if(pInstructionTable[i] != '') {
			tmpInstructionSet = tmpInstructionSet + pInstructionTable[i] + '|'; // Concat non-empty instructions with '|'
		}
	}
	if(tmpInstructionSet.length > 0) {
		tmpInstructionSet = tmpInstructionSet.slice(0, -1); // Remove trailing '|'
		tmpInstructionSet = 'C!W!' + tmpInstructionSet;     // Prefix with the instructionSet canceller (cancels any previous instructionChain)
	} else {
		tmpInstructionSet = emptyInstructionSet;            // Empty instructionSet (stop all motors)
	}
	return tmpInstructionSet;
}

// Ready function - this is exectuted once the web page is fully loaded
$(document).ready(function(){

	// Append version number to some elements
	$('title').text($('title').text() + ' v' + version);
	$('body h1').text($('body h1').text() + ' v' + version);
	
	// Hide all colored layers of the robot picture
	$('div#robot img.img-robot-layer').hide();
	
	// Hide all colored arrow buttons
	$('div#control img.img-arrow-layer').hide();
	
	// Initialize the URL
	$('#url').val(baseURL);
	
	// Initialize the status and buttons
	updateStatusAndButtons(false);
	
	// Initialize the replay script
	$('#recap').text('C!W!A,S,H,0,1');

	// Connect to Youba	
	$('#btnConnect').click(function(){
		var urlValue = $('#url').val();
		websocket = new WebSocket(urlValue);
		websocket.onopen = function(evt) { wsOnOpen(evt) };
		websocket.onclose = function(evt) { wsOnClose(evt) };
		websocket.onmessage = function(evt) { wsOnMessage(evt) };
		websocket.onerror = function(evt) { wsOnError(evt) };
	});
	
	// Disconnect from Youba		
	$('#btnDisconnect').click(function(){
		if(websocket != null) {
			websocket.close();
		}
	});
	
	// Display credits
	$('#btnAbout').click(function(){
		alert('Youga v' + version + '\nCopyright (C) 2017 AlphaK - http://www.alphak.net/');
	});
	
	// Redefine some of the gamepad API variables and functions
	gamepadSupport.tick = function() {
		gamepadSupport.pollStatus();
		updateInstructionSetObjectAndGui(); // <-- Hook from gamepad API to our implementation
		gamepadSupport.scheduleNextTick();
	};
	
	// Check and load settings from configuration file
	var configParameters = ['backToInitial', 'appendPosition', 'replayScript', 'baseLeft', 'baseRight', 'shoulderDown', 'shoulderUp', 'elbowDown', 'elbowUp', 'wristDown', 'wristUp', 'handLeft', 'handRight', 'pliersTight', 'pliersLoose'];
	var tmpConfigOK = true;
	for (var configIterator = 0 ; configIterator < configParameters.length ; configIterator++) {
		// Check one setting at a given time
		configIteratorValue = configParameters[configIterator];
		if (window[configIteratorValue] !== undefined
			&& Array.isArray(window[configIteratorValue]) 
			&& window[configIteratorValue].length == 3
			&& [A, B].indexOf(window[configIteratorValue][0]) != -1
			&& Number.isInteger(window[configIteratorValue][1])
			&& [0, 1].indexOf(window[configIteratorValue][2]) != -1
		) { 
			// Setting is correct: populate gamepadMap variable with setting values
			gamepadMap.push({
				'type': window[configIteratorValue][0],
				'index': window[configIteratorValue][1],
				'sign': window[configIteratorValue][2],
				'human': configIteratorValue, 
				'value': 0, 
				'youbaMotor': configMap[configIteratorValue][0],
				'youbaDir': configMap[configIteratorValue][1],
				'youbaSpecial': configMap[configIteratorValue][2]
			});
		}
		else if (window[configIteratorValue] !== undefined && window[configIteratorValue] === null) {
			// Setting is correct: not defined by user on purpose
		}
		else {
			// Setting is incorrect: notify user and prevent from sending commands to Youba
			tmpConfigOK = false;
			configOK = false;
			$('#config-status').text('Config file status: error - check setting "'+ configIteratorValue + '"');
			break;
		}
	}
	if(tmpConfigOK) {
		// All settings are correct: notify user and allow sending commands to Youba
		$('#config-status').text('Config file status: OK');
		configOK = true;
	}
	
});
